
f <- function(net) {
	ret = 1.0 / (1.0 + exp(-net))
	return (ret)
}

mlp.architecture <- function(input.layer.size = 2,
		hidden.layer.size = 2,
		output.layer.size = 1,
		f.net = f) {

	layers = list()
	layers$hidden = matrix(runif(min=-1, max=1, 
				  n=hidden.layer.size*(input.layer.size+1)), 
			       nrow=hidden.layer.size, 
			       ncol=input.layer.size+1)
	#		w1	w2	theta
	# neuron 0	0.01   -0.9     -0.7
	# neuron 1	0.78	0.34	-0.3
	#
	layers$output = matrix(runif(min=-1, max=1, 
				  n=output.layer.size*(hidden.layer.size+1)), 
			       nrow=output.layer.size, 
			       ncol=hidden.layer.size+1)
	#		w1	w2	theta
	# neuron 0	0.01   -0.9     -0.7
	#

	ret = list()
	ret$input.layer.size = input.layer.size
	ret$hidden.layer.size = hidden.layer.size
	ret$output.layer.size = output.layer.size
	ret$layers = layers
	ret$f.net = f.net

	return (ret)
}

forward <- function(architecture, dataset, p) {
	# apenas produzir a saida

	# xor.dataset
	# 0 0 0
	# 0 1 1
	# 1 0 1
	# 1 1 0

	X = matrix(dataset[,1:architecture$input.layer.size], ncol=architecture$input.layer.size)
	Y = matrix(dataset[,(architecture$input.layer.size+1):ncol(dataset)], nrow=nrow(X))

	# Aplicar as entradas X na camada escondida
	net_h = architecture$layers$hidden %*% c(as.vector(ts(X[p,])), 1)
	f_net_h = architecture$f.net(net_h)

	# Aplicar as saidas produzidas pela camada escondida
	#	na camada de saida
	net_o = architecture$layers$output %*% c(f_net_h, 1)
	f_net_o = architecture$f.net(net_o)

	ret = list()
	ret$f_net_h = f_net_h
	ret$f_net_o = f_net_o

	return (ret)
}

# Treinamento (adaptação de pesos)
backpropagation <- function(architecture, dataset, eta=0.1, threshold=1e-3) {
	# alterar pesos w e theta

	X = matrix(dataset[,1:architecture$input.layer.size], ncol=architecture$input.layer.size)
	Y = matrix(dataset[,(architecture$input.layer.size+1):ncol(dataset)], nrow=nrow(X))

	cat("Input data...\n")
	print(X)

	cat("Expected output...\n")
	print(Y)

	cat("Enter to start running...")
	readline()

	squared_error = threshold * 2

	while (squared_error > threshold) {
		squared_error = 0
		for (p in 1:nrow(X)) {
			# aplicar na mlp a entrada
			f = forward(architecture, dataset, p)

			# delta de output
			error = (Y[p,] - f$f_net_o)
			delta_o = error * f$f_net_o * (1-f$f_net_o)

			# sum porque podemos ter mais de 1 neurônio na camada de saída
			squared_error = squared_error + sum(error^2) 

			# delta de hidden
			w_o = architecture$layers$output[,1:architecture$hidden.layer.size]
			delta_h = (f$f_net_h * (1 - f$f_net_h)) * 
				sum(as.vector(delta_o) * as.vector(w_o))

			# aprender
			architecture$layers$output = 
				architecture$layers$output + eta * 
					delta_o %*% c(f$f_net_h, 1)

			architecture$layers$hidden =
				architecture$layers$hidden + eta * 
					delta_h %*% c(X[p,], 1)
		}

		squared_error = squared_error / nrow(X)

		cat("Squared error = ", squared_error, "\n")
	}
	
	return (architecture)
}

# Para executar a MLP depois de treinada
mlp.test <- function(architecture, dataset, debug=T) {

	X = matrix(dataset[,1:architecture$input.layer.size], ncol=architecture$input.layer.size)
	Y = matrix(dataset[,(architecture$input.layer.size+1):ncol(dataset)], nrow=nrow(X))

	cat("Enter to start testing...")
	readline()

	output = NULL

	for (p in 1:nrow(X)) {
		# aplicar na mlp a entrada
		f = forward(architecture, dataset, p)

		if (debug) {
			cat("Input pattern = ", as.vector(X[p,]), 
			    " Expected output = ", as.vector(Y[p,]), 
				" Obtained output = ", as.vector(f$f_net_o), "\n")
		}

		output = rbind(output, as.vector(f$f_net_o))
	}

	return (output)
}

discretize.hiperplane <- function(img, range = c(0.45, 0.55)) {
	ids_negative = which(img < range[1])
	ids_positive = which(img > range[2])
	ids_hiperplane = which(img >= range[1] & img <= range[2])

	img[ids_negative] = 0
	img[ids_positive] = 1
	img[ids_hiperplane] = 0.5

	img
}

xor.test <- function(eta=0.25, threshold=1e-3) {
	dataset = as.matrix(read.table("xor.dat"))

	model = mlp.architecture(input.layer.size = 2, 
				 hidden.layer.size = 2, 
				 output.layer.size = 1, f.net = f)

	trained.model = backpropagation(model, 
					dataset, 
					eta=eta, 
					threshold=threshold)

	mlp.test(trained.model, dataset)

	X = seq(-0.1,1.1,length=100)
	hiperplane_1 = outer(X,X, function(X,Y) { cbind(X,Y,1) %*% trained.model$layers$hidden[1,] } )
	hiperplane_2 = outer(X,X, function(X,Y) { cbind(X,Y,1) %*% trained.model$layers$hidden[2,] } )

	cat("Press enter to plot both hiperplanes...")
	readline()

	#par(mfrow=c(1,2))
	filled.contour(discretize.hiperplane(hiperplane_1) + 
		       discretize.hiperplane(hiperplane_2))
}

ocr.test <- function(eta=0.25, threshold=1e-3) {
	train.dataset = as.matrix(read.table("ocr-as.vector.dat"))
	test.dataset = as.matrix(read.table("test.ocr-as.vector.dat"))

	model = mlp.architecture(input.layer.size = 70, 
				 hidden.layer.size = 3, 
				 output.layer.size = 2, f.net = f)

	trained.model = backpropagation(model, 
					train.dataset, 
					eta=eta, 
					threshold=threshold)
	mlp.test(trained.model, test.dataset)
}

wine.test <- function(eta=0.25, threshold=1e-3) {
	train.dataset = as.matrix(read.table("wine.train"))
	test.dataset = as.matrix(read.table("wine.test"))

	model = mlp.architecture(input.layer.size = 13, 
				 hidden.layer.size = 5, 
				 output.layer.size = 3, f.net = f)

	trained.model = backpropagation(model, 
					train.dataset, 
					eta=eta, 
					threshold=threshold)
	mlp.test(trained.model, test.dataset)
}

